static char *cvs_id =" $Id: xnd_cplx.c,v 1.2 2002/01/10 15:24:43 berar Exp $ ";
/* XND_cplx.C  complex functions used in xnd                  */
/* ---------------------------------------------------------- */
/* based on public domain num recipes and                     */
/* GAMS from NIST Comp. Sc. Div., class C4a, pack CMLIB       */
/* ---------------------------------------------------------- */
#include <math.h>                /* double fabs +... sin, cos */
/* ---------------------------------------------------------- */
#include "../xnd_def.h"
#include "xnd_t_cplx.h"
/* ---------------------------------------------------------- */
/* ---------------------------------------------------------- */
fcomplex Cadd(fcomplex a, fcomplex b)
{
	fcomplex c;
	c.r=a.r+b.r;
	c.i=a.i+b.i;
	return c;
}
/* ---------------------------------------------------------- */

fcomplex Csub(fcomplex a, fcomplex b)
{
	fcomplex c;
	c.r=a.r-b.r;
	c.i=a.i-b.i;
	return c;
}
/* ---------------------------------------------------------- */
fcomplex Cmul(fcomplex a, fcomplex b)
{
	fcomplex c;
	c.r=a.r*b.r-a.i*b.i;
	c.i=a.i*b.r+a.r*b.i;
	return c;
}
/* ---------------------------------------------------------- */
fcomplex Complex(float re, float im)
{
	fcomplex c;
	c.r=re;
	c.i=im;
	return c;
}
/* ---------------------------------------------------------- */
fcomplex Conjg(fcomplex z)
{
	fcomplex c;
	c.r=z.r;
	c.i = -z.i;
	return c;
}
/* ---------------------------------------------------------- */
fcomplex Cdiv(fcomplex a, fcomplex b)
{
	fcomplex c;
	float r,den;
	if (fabs(b.r) >= fabs(b.i)) {
		r=b.i/b.r;
		den=b.r+r*b.i;
		c.r=(a.r+r*a.i)/den;
		c.i=(a.i-r*a.r)/den;
	} else {
		r=b.r/b.i;
		den=b.i+r*b.r;
		c.r=(a.r*r+a.i)/den;
		c.i=(a.i*r-a.r)/den;
	}
	return c;
}
/* ---------------------------------------------------------- */
float Cabs(fcomplex z)
{
	float x,y,ans,temp;
	x=fabs(z.r);
	y=fabs(z.i);
	if (x == 0.0)
		ans=y;
	else if (y == 0.0)
		ans=x;
	else if (x > y) {
		temp=y/x;
		ans=x*sqrt(1.0+temp*temp);
	} else {
		temp=x/y;
		ans=y*sqrt(1.0+temp*temp);
	}
	return ans;
}
/* ---------------------------------------------------------- */
fcomplex Csqrt(fcomplex z)
{
	fcomplex c;
	float x,y,w,r;
	if ((z.r == 0.0) && (z.i == 0.0)) {
		c.r=0.0;
		c.i=0.0;
		return c;
	} else {
		x=fabs(z.r);
		y=fabs(z.i);
		if (x >= y) {
			r=y/x;
			w=sqrt(x)*sqrt(0.5*(1.0+sqrt(1.0+r*r)));
		} else {
			r=x/y;
			w=sqrt(y)*sqrt(0.5*(r+sqrt(1.0+r*r)));
		}
		if (z.r >= 0.0) {
			c.r=w;
			c.i=z.i/(2.0*w);
		} else {
			c.i=(z.i >= 0) ? w : -w;
			c.r=z.i/(2.0*c.i);
		}
		return c;
	}
}
/* ---------------------------------------------------------- */
fcomplex RCmul(float x, fcomplex a)
{
	fcomplex c;
	c.r=x*a.r;
	c.i=x*a.i;
	return c;
}
/* ---------------------------------------------------------- */
/* functions adapted from C4a GAMS/NIST by Fullerton  820801  */
/*  c equiv to fortran R1mach constants :                     */
/*  R1mach(1)  MINFLOAT        R1mach(2)  MAXFLOAT            */
/*  R1mach(3)  2^-FSIGNIF      R1mach(4)  2^(1-FSIGNIF)       */
static int nterms = 0;
static float sqeps;
static float rmin;
static float rmax;
static void iniCatanCtan(void)
{  double r1mach3 = pow(2.0, -FSIGNIF);
	nterms = -0.4343*log(r1mach3) + 1.0;
   sqeps = sqrt(2*r1mach3);
   rmin = sqrt (3.0*r1mach3);
   rmax = 1/r1mach3;
	}
static complex_error = 10;
static complex_warning = 5;
/* ---------------------------------------------------------- */
fcomplex Catan(fcomplex z)
/* adapted from  CATAN, C4a GAMS/NIST by Fullerton  820801    */
{	fcomplex zret, z2, zz;
	float  r, r2;
	int twoi, i;
	
	if ( nterms==0 ) iniCatanCtan();
   r =  Cabs(z);
	if (r <= 0.1 )
	{  zret = z;
      if (r < rmin ) return zret;

      zret.r = 0.0;
		zret.i = 0.0;
      z2 = Cmul(z,z);
      twoi = 2*nterms + 1;
		for( i=0; i<nterms; i++)
		{  twoi -= 2;
         zz = Cmul(z2,zret);
			zret.r = 1.0/twoi - zz.r;
			zret.i = - zz.i;
			}
      zret = Cmul(z,zret);
      return zret;
      }
	if (r <= rmax )
   {  r2 = r*r;
      if (r2==1.0 && z.r == 0.0) 
		{  if (complex_error-->0) fprintf(stderr,
		        " Catan(%f,%f) : singular z == i\n", z.r,z.i);
	      }
      if (fabs(r2-1.0) <= sqeps) 
	   {  if (complex_warning-->0) fprintf(stderr, 
	            " Ctan(%f,%f) : answer < half precision, z^2 close to -1\n", z.r,z.i);
	      }
      zret.r = 0.5*atan2(2.0*z.r, 1.0-r2);
      zret.i = 0.25*log((r2+2.0*z.i+1.0)/(r2-2.0*z.i+1.0));
      return zret;
      }
   zret.r = z.r < 0 ? -M_PI_2 : M_PI_2;
   zret.i = 0;
   return zret;
   }
/* ---------------------------------------------------------- */
fcomplex Catan2(fcomplex csn,fcomplex ccs)
/* adapted from  CATAN2, C4a GAMS/NIST by Fullerton  820801   */
{  fcomplex zret;
   if (Cabs(ccs)!=0) 
	{  zret = Catan(Cdiv(csn,ccs));
	   if (ccs.r < 0) zret.r += M_PI;
		if (zret.r > M_PI ) zret.r -= 2 * M_PI;
      return zret;
      }
   if (Cabs(ccs)!=0)
	{  if (complex_error-->0) fprintf(stderr, 
	           " Catan2(%f,%f) both args are zero\n", csn, ccs);
	      }
   zret.r = csn.r < 0 ? -M_PI_2 : M_PI_2;
   zret.i = 0;
   return zret;
   }
/* ---------------------------------------------------------- */
fcomplex Catanh(fcomplex z)
/* adapted from  CATAN2, C4a GAMS/NIST by Fullerton  820801   */
{   
/*	return -Complex(0,1)*Catan(Cmul(Complex(0,1),z)); */
	return Cmul(Complex(0,-1),Catan(Cmul(Complex(0,1),z)));
	}
/* ---------------------------------------------------------- */
fcomplex Ctan(fcomplex z)
/* adapted from  CATAN2, C4a GAMS/NIST by Fullerton  820801   */
{  fcomplex zret;
	fcomplex z2;
	float den, sin2x;
	
	if ( nterms==0 ) iniCatanCtan();
	z2 = RCmul(2.0, z);
   sin2x = sin(z2.r);
	den = cos(z2.r) + cosh(z2.i);
	if (den==0)	
	{  if (complex_error-->0) fprintf(stderr, 
	    " Ctan(%f,%f) : singular Pi/2 and z.i==0\n", z.r,z.i);
	   }
   if (fabs(den) <= sqeps || fabs(den) <= z2.r*sqeps)
	{  if (complex_warning-->0) fprintf(stderr, 
	    " Ctan(%f,%f) : answer < half precision, abs(z.r) too big or z too near Pi/2\n", z.r,z.i);
	  }
	zret.r = sin2x/den;
	zret.i = sinh(z2.i)/den;
	return zret;
	}
/* ---------------------------------------------------------- */
fcomplex Ctanh(fcomplex z)
/* adapted from  CATAN2, C4a GAMS/NIST by Fullerton  820801   */
{   
	return Cmul(Complex(0,-1),Ctan(Cmul(Complex(0,1),z)));
/*	return -Complex(0,1)*Ctan(Cmul(Complex(0,1),z)); */
	}
/* ---------------------------------------------------------- */

#
