static char *cvs_id =" $Id: xnd_num.c,v 1.1.1.1 1999/08/24 07:38:00 berar Exp $ ";
/* XND_NUM.C   numerical function used in xnd                 */
/*        --------------------------------                    */                                     
/* ---------------------------------------------------------- */
/* bessel_j functions are adapted from numerical recipes      */
/* results checked with mathematica from 0 to 6.0 11Nov98     */
/* ---------------------------------------------------------- */
#include <math.h>                /* double fabs +... sin, cos */
#include "xnd_nrh.h"
/* ---------------------------------------------------------- */
/* ---------------------------------------------------------- */
/* bessel_j0, bessel_j1 are called by the common dispatcher   */
/* ---------------------------------------------------------- */
float bessel_j0( float x)
{  double ax, xx, y, z, ans, ans1, ans2;
	ax=fabs(x);
   if (ax < 8.0 )
   {   y = ax*ax;
       ans1=57568490574.0+y*(-13362590354.0+y*(651619640.7
           +y*(-11214424.18+y*(77392.33017+y*(-184.9052456)))));
       ans2=57568490411.0+y*(1029532985.0+y*(9494680.718
           +y*(59272.64853+y*(267.8532712+y))));
       ans = ans1/ans2;
       return ans1/ans2;
       }   
   z = 8.0/ax;
   y=z*z;
   xx=ax-0.785398164;
   ans1=1.0+y*(-0.1098628627e-2+y*(0.2734510407e-4
       +y*(-0.2073370639e-5+y*(0.2093887211e-6))));
   ans2=-0.1562499995e-1+y*(0.1430488765e-3
       +y*(-0.6911147651e-5+y*(0.7621095161e-6
       -y*0.934935152e-7)));
   ans = sqrt(0.636619772/ax)*cos(xx)*ans1-z*sin(xx)*ans2;
   return ans;
   }   
/* ---------------------------------------------------------- */
float bessel_j1( float x)
{  double ax, xx, y, z, ans, ans1, ans2;
	ax=fabs(x);
   if (ax < 8.0 )
   {  y = ax*ax;
      ans1=ax*(72362614232.0+y*(-7895059235.0+y*(242396853.1
           +y*(-2972611.439+y*(15704.48260+y*(-30.16036606))))));
      ans2=144725228442.0+y*(2300535178.0+y*(18583304.74
           +y*(99447.43394+y*(376.9991397+y))));
		ans = ans1/ans2;
      return ans1/ans2;
      }
    z = 8.0/ax;
    y=z*z;
    xx=ax-2.356194491;
    ans1=1.0+y*(0.183105e-2+y*(-0.3516396496e-4
       +y*(0.2457520174e-5+y*(-0.240337019e-6))));
    ans2=0.4687499995e-1+y*(-0.2002690873e-3
       +y*(0.8449199096e-5+y*(-0.88228987e-6
       +y*0.105787412e-6)));
    ans = sqrt(0.636619772/ax)*cos(xx)*ans1-z*sin(xx)*ans2;
    return ans;
    } 
/* ---------------------------------------------------------- */
/* ---------------------------------------------------------- */
#define ACC 40.0
#define BIGNO   1.0e10
#define BIGNI   1.0e-10
float bessel_j (int n, float x)
{  int j, jsum, m, sgn;
   float ax, bj,bjm,bjp,sum,tox,ans;
	switch (n) 
   {  case 0 :  return  bessel_j0(x);   
      case 1 :  return  bessel_j1(x);   
      case -1 : return -bessel_j1(x);   
      default :
      sgn=0;
	   if ( n < 0 ) {n = -n;  if (n&1) sgn=1; } 
      ax = fabs(x);
	   if (ax == 0 ) return 0;
	   if ( ax > (float) n)
	   {  tox = 2.0/ax;
	      bjm = bessel_j0(ax);
	      bj  = bessel_j1(ax);
	      for ( j=1; j<n; j++)
	      {  bjp = j*tox*bj-bjm;
	         bjm = bj;
		      bj = bjp;
		      }
	      ans = bj;
	      } 
		else
	   {  tox = 2.0/ax;
	      m=2*((n+(int)sqrt(ACC*n))/2);
	      jsum =0;
	      bjp=ans=sum=0;
	      bj  = 1.0;
	      for ( j=m; j>0; j--)
	      {  bjm = j*tox*bj-bjp;
	         bjp = bj;
		      bj = bjm;
		      if ( fabs(bj) > BIGNO)
            {  bj *= BIGNI;
		         bjp *=BIGNI;
		         ans *= BIGNI;
		         sum *= BIGNI;
		         }
	         if (jsum) sum += bj;
	         jsum = !jsum;
	         if (j == n) ans = bjp;
	         } 
	      sum = 2.0*sum-bj;
	      ans /= sum;
	      } 
	   if ( x < 0.0 && ( n & 1 )) ans = -ans;
		return ( sgn ) ? -ans : ans;       
     }
    }
/* ---------------------------------------------------------- */
/* derivatives from BesselJ according mathematica             */
float dbessel_j (int n, float x)
{  switch (n) 
   {  case 0 : return -bessel_j1(x);   
      case 1 : return 0.5*(bessel_j0(x)-bessel_j(2,x));   
      case 2 : return 0.5*(bessel_j1(x)-bessel_j(n+1,x));   
      default: return 0.5*(bessel_j(n-1,x)-bessel_j(n+1,x));
      }
   }
/* ---------------------------------------------------------- */
/* ---------------------------------------------------------- */
