static char *cvs_id =" $Id: xnd_swap.c,v 1.2 2002/01/10 15:24:45 berar Exp $ ";
/* xnd_swap.c     functions loading data for each scan    */               
/* ------------------------------------------------------------ */
/* internal memory handler used for swapping in the derivatives */
/* table if needed                                              */
/* read and write also the binary data file                     */
/* ------------------------------------------------------------ */
#define USE_SCAN
#define USE_OBS
#include "../xnd_def.h"
#include <string.h>
#ifdef __TURBOC__
#include <alloc.h>
#endif
#ifndef SEEK_SET
#define SEEK_SET 0
#endif

#include "../xnd_mem.h"

#ifdef XND_SITE
#undef ftell
#undef fseek
#endif
/* ------------------------------------------------------------ */
/* routines allowing i_obs, weight, i_calc, ...  and the table  */
/* of the derivatives dy_dp to be stored on disk                */
/* they are totally contained in memory if disk_dydp is zero   */
static FLOAT  *dy_dp;          /* derivative table */
static unsigned dydp_nb;       /* number of points in memory    */
static unsigned dydp_size;     /* size allocated for each point */
static unsigned dydp_dim;
static unsigned dydp_pos;      /* first point in memory */
static unsigned dydp_max;      /* above point limit   */
static unsigned dydp_mod;      /* modified memory flag */
static FILE *   dydp_file = NULL;   /* memory file */
static FILE *   data_file = NULL;   /* powder data file */
static int data_file_mode;
/* ------------------------------------------------------------ */
FLOAT (* get_dydp_val)(unsigned pts, unsigned pos);
void  (* set_dydp_val)(unsigned pts, unsigned pos, FLOAT f);
void  (* add_dydp_val)(unsigned pts, unsigned pos, FLOAT f);
/* ------------------------------------------------------------ */
/* ------------------------------------------------------------ */
/*  function used when all is in memory                         */
/* ------------------------------------------------------------ */
static FLOAT  memory_get_val(unsigned pts, unsigned pos)
{   return dy_dp[pts*dydp_dim + pos];
    }
/* ------------------------------------------------------------ */
static void  memory_set_val(unsigned pts, unsigned pos, FLOAT f)
{   dy_dp[pts*dydp_dim+pos] = f;
    }
/* ------------------------------------------------------------ */
static void  memory_add_val(unsigned pts, unsigned pos, FLOAT f)
{   dy_dp[pts*dydp_dim+pos] += f;
    }
/* ------------------------------------------------------------ */
/* ------------------------------------------------------------ */
/*  function used when some swapping is used                    */
static void rw_dydp (P_(unsigned) pos);
/* ------------------------------------------------------------ */
static FLOAT  swap_get_val( unsigned pts, unsigned pos)
{   if (pts < dydp_pos || pts >= dydp_max) rw_dydp (pts);
    return dy_dp[ (pts-dydp_pos)*dydp_dim + pos ];
    }
/* ------------------------------------------------------------ */
static void  swap_set_val(unsigned pts, unsigned pos, FLOAT f)
{   if (pts < dydp_pos || pts >= dydp_max) rw_dydp (pts);
    dydp_mod = 1;
    dy_dp[(pts-dydp_pos)*dydp_dim + pos] = f;
    }
/* ------------------------------------------------------------ */
static void  swap_add_val(unsigned pts, unsigned pos, FLOAT f)
{   if (pts < dydp_pos || pts >= dydp_max) rw_dydp (pts);
    dydp_mod = 1;
    dy_dp[(pts-dydp_pos)*dydp_dim + pos] += f;
    }
/* ------------------------------------------------------------ */
/* used at the beginning of each scan to select functions       */
/* ------------------------------------------------------------ */
void memory_new (P_(int) n_par, P_(int) tot_phase)
{    int  i,j,n,m;
     TYPE_OBS  obs;
     FLOAT *point;
     FLOAT *p_o;
     push_err_msg("memory_new(xnd_swap.c)");
     n_pts = l_diag->n_points;
     point = l_diag->points = (FLOAT *)tst_alloc(n_pts, sizeof(FLOAT),"points");
     dydp_nb = disk_dydp > n_pts ? n_pts : disk_dydp;
     dydp_dim = DYDP_IPHA + 1;
     dydp_dim += ( tot_phase >= n_par ) ? tot_phase : n_par;
     dydp_size = dydp_dim * sizeof(FLOAT);
     dy_dp = (FLOAT *) tst_alloc (dydp_nb, dydp_size, "dy_dp");
     memset (dy_dp,0, dydp_nb * dydp_size);
     dydp_mod = 0;
     if (dydp_file) if (fseek (dydp_file, 0, SEEK_SET)!=0) termine (15,E_DYDPS);
     m = 1+(n_pts-1) / dydp_nb;
	  for (i=0, n=0; i<m; i++)
     {  for (j=0,p_o = dy_dp; j<dydp_nb; j++,n++,p_o+=dydp_dim)
	     if (n<n_pts)
	     {   data_read (&obs);
	         *point++ = obs.dtheta;
	         *(p_o+DYDP_IOBS) = obs.i_obs;
	         *(p_o+DYDP_WGTH) = obs.weight;
	         } else
	     {   memset(p_o,0, dydp_size);
	         }
	    if (dydp_file) if (dydp_nb!=fwrite (dy_dp, dydp_size, dydp_nb, dydp_file))
	      termine (15,E_DYDPW);
	    }
    dydp_max = i*dydp_nb;
    dydp_pos = dydp_max - dydp_nb;
    if (dydp_file)
	 {   get_dydp_val = swap_get_val;
	     set_dydp_val = swap_set_val;
	     add_dydp_val = swap_add_val;
	     }
	 else
	 {   get_dydp_val = memory_get_val;
	     set_dydp_val = memory_set_val;
	     add_dydp_val = memory_add_val;
	     }
    pop_err_msg();
	 }
/* ------------------------------------------------------------ */
void memory_free (P_(void))    /* save previous ic */
{  int i;
   TYPE_OBS  obs;
   push_err_msg("memory_free(xnd_swap.c)");
   data_read (&obs);         /* check value if not temporary */
   data_read (&obs);  
   pop_err_msg();
   push_err_msg("memory_free : l_diag (xnd_swap.c)");
   free (l_diag->points);
   pop_err_msg();
   push_err_msg("memory_free : dy_dp (xnd_swap.c)");
   free (dy_dp);
   pop_err_msg();
	}
/* ------------------------------------------------------------ */
long memory_swap (int max_points, int n_par, int lect_disk_dydp)
{	long   l, f_size;
   int dydp_dim;
   push_err_msg("memory_swap (xnd_swap.c)");
   dydp_dim = DYDP_IDER + 1;
   dydp_dim += (tot_phase >= n_par) ? tot_phase : n_par;
   dydp_file = NULL;
   f_size = max_points * dydp_dim * sizeof(FLOAT);
   disk_dydp = 0;
	if (MAX_ALLOC)  /* check if not zero or undefined */
   {  l = floor ( MAX_ALLOC / (FLOAT) (dydp_dim*sizeof(FLOAT)));
		if (lect_disk_dydp == 0)
		{   if (l > max_points)
			 {   disk_dydp = max_points;
				  fprintf (list_fich,"disk_dydp parameter : %d points are in memory\n", disk_dydp);
				  f_size = 0;
              }
			 else
			 {   disk_dydp =  (l > 100) ? 20 : l ;
				  fprintf (list_fich," parametre disk_dydp defini : %d points en memoire, maximum %ld\n",disk_dydp,l);
			 }   }
		else if ( lect_disk_dydp > l)
		{   disk_dydp = l;
			 fprintf (list_fich," parametre disk_dydp modifie : %d points en memoire\n",disk_dydp);
			 }
      else
      {   f_size = 0;
          fprintf (list_fich,"disk_dydp parameter : %d points are in memory (max = %ld)\n", disk_dydp,l);
		    disk_dydp = max_points;
	   }   }
    fprintf (list_fich," Size of temporary derivative array : %ld ko\n",
    1L+((1L+max_points)*dydp_dim)/1024);
    pop_err_msg();
	 return f_size;
    }
/* ------------------------------------------------------------ */
void memory_adjust(int n_par)
{	long   l, f_size;
   int dydp_dim;
   dydp_dim = DYDP_IDER + 1 + (tot_phase >= n_par) ? tot_phase : n_par;
   fprintf(list_fich, "Sizes have been modified for %d parameters\n", n_par);
	}
/* ------------------------------------------------------------ */
void memory_file (FILE * tmp)
{   dydp_file = tmp;
    tst_abort ((dydp_file==NULL),"swap file");
    }
/* ------------------------------------------------------------ */
void memory_close (void)
{   if (dydp_file) fclose(dydp_file);
    dydp_file = NULL;
    }
/* ------------------------------------------------------------ */
static void rw_dydp (P_(unsigned) pos)
#ifdef NO_PROTO
unsigned pos;
#endif
{  unsigned long offset;
   if (dydp_mod)
   {  offset =  dydp_pos * (long) dydp_size;
	   if (fseek (dydp_file, offset, SEEK_SET)!=0)
	 	termine (15,E_DYDPS);
	   if (fwrite (dy_dp, dydp_size, dydp_nb, dydp_file)!=dydp_nb)
	 	termine (15,E_DYDPW);
	   dydp_mod = 0;
	   }
   dydp_pos = (pos / dydp_nb) * dydp_nb;
   dydp_max = dydp_nb + dydp_pos;
   offset =  dydp_pos * (long) dydp_size;
   if (fseek (dydp_file, offset, SEEK_SET)!=0)
   termine (15,E_DYDPS);
   if (fread (dy_dp, dydp_size, dydp_nb, dydp_file)!=dydp_nb) 
   termine (15,E_DYDPR);
   }
/* ------------------------------------------------------------ */
/* --------- end of the derivatives table routines ------------- */
/* ------------------------------------------------------------ */
void data_read (P_(TYPE_OBS *) obs)
#ifdef NO_PROTO
TYPE_OBS * obs;
#endif
{   if (fread ( (void *) obs,   sizeof(TYPE_OBS),  1, data_file) != 1)
       termine (15,E_RDATF);
    }
/* ------------------------------------------------------------ */
void data_write (P_(TYPE_OBS *) obs)
#ifdef NO_PROTO
TYPE_OBS * obs;
#endif
{   if (fwrite ((void *) obs, sizeof(TYPE_OBS), 1, data_file) != 1)
       termine (15,E_WDATF);
    }
/* ------------------------------------------------------------ */
void data_init( FILE * f, int file_mode)
{   data_file = f;
    tst_abort ((f==NULL)," bin data file");
    data_file_mode = file_mode;
    }
/* ------------------------------------------------------------ */
void data_start( void )
{    rewind (data_file);
    }
/* ------------------------------------------------------------ */
void data_close( void )
{   if ( data_file ) fclose (data_file);
    data_file= NULL;
    }
/* ------------------------------------------------------------ */
/* ------------------------------------------------------------ */

